"""This module provides miscellaneous tools"""

__all__ = ["IsNotebook", "animate_init_", "animate_next_", "ReferenceVectors"]

from desdeo_emo.othertools.IsNotebook import IsNotebook
from desdeo_emo.othertools.plotlyanimate import animate_init_, animate_next_
from desdeo_emo.othertools.ReferenceVectors import ReferenceVectors
